#include "terminal_api_driver.h"

#include <cmath>
#include <sstream>
#include <iomanip>
#include <shared_mutex>
#include <boost/thread.hpp>
#include <boost/log/trivial.hpp>
#include <boost/property_tree/json_parser.hpp>
#include <sys/sysinfo.h>
#include "common/nlohmann/json.hpp"
#include "api-driver/daemon.h"

typedef boost::property_tree::ptree::path_type json_path;


api_driver::ApiDriver::ApiDriver() = default;

void api_driver::ApiDriver::startDaemon() {
    if (daemon == nullptr) {
        daemon = std::make_unique<TerminalApiDaemon>();
        BOOST_LOG_TRIVIAL(info) << "api_driver::ApiDriver::startDaemon(): API daemon succes started!";
    }
}

std::string api_driver::buildEscapedString(const std::string& source) {
    std::string str(source);
    size_t start_pos = 0;
    while((start_pos = str.find('\"', start_pos)) != std::string::npos) {
        str.replace(start_pos, 1, "\\\"");
        start_pos += 2;
    }
    for (start_pos = 0; start_pos < str.size() && (str[start_pos] == ' ' || str[start_pos] == '\n' || str[start_pos] == '\t'); start_pos++) {}
    size_t end_pos = str.size() - 1;
    for (; end_pos > start_pos && end_pos != 0 && (str[end_pos] == ' ' || str[end_pos] == '\n' || str[end_pos] == '\t'); end_pos--) {}
    return "\"" + str.substr(start_pos, end_pos - start_pos + 1) + "\"";
}

std::tuple<uint8_t, uint8_t> translateCoordinates(double abs) {
    auto deg = static_cast<uint8_t>(abs);
    double min_double = (abs - deg) * 60;
    auto min = static_cast<uint8_t>(min_double);
    return std::make_tuple(deg, min);
}


std::string api_driver::ApiDriver::loadTerminalState() const {
    if (daemon == nullptr) {
        return R"({"error": "api daemon not started!"})";
    }

    obj::TerminalState state;
    daemon->getState(state);
    obj::TerminalDeviceState devState;
    daemon->getDeviceState(devState);

    auto res = state.asJson();
    res["device"] = devState.asJson();
    return res.dump();
}


void api_driver::ApiDriver::resetPacketStatistics() const {
    this->daemon->resetPacketStatistics();
}

#ifdef MODEM_IS_SCPC
struct ModcodDef_t {const char* modulation; const char* speed;};
const static ModcodDef_t ModcodDefs[] = {
    {.modulation = "dummy", .speed = "0"},
    {.modulation = "qpsk", .speed = "1/4"},
    {.modulation = "qpsk", .speed = "1/3"},
    {.modulation = "qpsk", .speed = "2/5"},
    {.modulation = "qpsk", .speed = "1/2"},
    {.modulation = "qpsk", .speed = "3/5"},
    {.modulation = "qpsk", .speed = "2/3"},
    {.modulation = "qpsk", .speed = "3/4"},
    {.modulation = "qpsk", .speed = "4/5"},
    {.modulation = "qpsk", .speed = "5/6"},
    {.modulation = "qpsk", .speed = "8/9"},
    {.modulation = "qpsk", .speed = "9/10"},
    {.modulation = "8psk", .speed = "3/5"},
    {.modulation = "8psk", .speed = "2/3"},
    {.modulation = "8psk", .speed = "3/4"},
    {.modulation = "8psk", .speed = "5/6"},
    {.modulation = "8psk", .speed = "8/9"},
    {.modulation = "8psk", .speed = "9/10"},
    {.modulation = "16apsk", .speed = "2/3"},
    {.modulation = "16apsk", .speed = "3/4"},
    {.modulation = "16apsk", .speed = "4/5"},
    {.modulation = "16apsk", .speed = "5/6"},
    {.modulation = "16apsk", .speed = "8/9"},
    {.modulation = "16apsk", .speed = "9/10"},
    {.modulation = "32apsk", .speed = "3/4"},
    {.modulation = "32apsk", .speed = "4/5"},
    {.modulation = "32apsk", .speed = "5/6"},
    {.modulation = "32apsk", .speed = "8/9"},
    {.modulation = "32apsk", .speed = "9/10"},
};

static const char* extractModcodModulation(uint32_t modcod, bool defaultQpsk1_4 = true) {
    modcod >>= 2;
    const auto* d = defaultQpsk1_4 ? ModcodDefs : ModcodDefs + 1;
    if (modcod < (sizeof(ModcodDefs) / sizeof(ModcodDef_t))) {
        d = ModcodDefs + modcod;
    }
    return d->modulation;
}

static const char* extractModcodSpeed(uint32_t modcod, bool defaultQpsk1_4 = true) {
    modcod >>= 2;
    const auto* d = defaultQpsk1_4 ? ModcodDefs : ModcodDefs + 1;
    if (modcod < (sizeof(ModcodDefs) / sizeof(ModcodDef_t))) {
        d = ModcodDefs + modcod;
    }
    return d->speed;
}
#endif

std::string api_driver::ApiDriver::loadSettings() const {
    if (daemon == nullptr) {
        return R"({"error": "api daemon not started!"})";
    }

    // TODO сделать получение настроек
    return "{}";
}

std::string api_driver::ApiDriver::loadFirmwareVersion() const {
    if (daemon == nullptr) {
        return R"({"error": "api daemon not started!"})";
    }

    std::stringstream result;
    auto firmware = daemon->getFirmware();
    result << "{\n\"fw.version\":" << buildEscapedString(firmware.version);
    result << ",\"fw.modemId\":" << buildEscapedString(firmware.modemId);
    result << ",\"fw.modemSn\":" << buildEscapedString(firmware.modemSn);
    result << ",\"fw.macMang\":" << buildEscapedString(firmware.macMang);
    result << ",\"fw.macData\":" << buildEscapedString(firmware.macData);
    result << "\n}";
    return result.str();
}

#ifdef MODEM_IS_SCPC
static uint32_t buildModcodFromPt(const boost::property_tree::ptree& pt, const std::string& name, bool isShortFrame, bool isPilots = false) {
    uint32_t modcod = 0;
    const auto mod = pt.get<std::string>(name + "Modulation");
    const auto speed = pt.get<std::string>(name + "Speed");
    uint32_t _index = 0;
    for (const auto& m: ModcodDefs) {
        if (mod == m.modulation) {
            if (modcod == 0) modcod = _index;
            if (speed == m.speed) {
                modcod = _index;
                break;
            }
        }
        _index++;
    }
    return (modcod << 2)| (isShortFrame ? 2 : 0) | (isPilots ? 1 : 0);
}
#endif

void api_driver::ApiDriver::setRxTxSettings(boost::property_tree::ptree &pt) {
    // TODO
}

void api_driver::ApiDriver::setDpdiSettings(boost::property_tree::ptree &pt) {
    // TODO
}

void api_driver::ApiDriver::setBucLnbSettings(boost::property_tree::ptree &pt) {
    // TODO
}

void api_driver::ApiDriver::setQosSettings(boost::property_tree::ptree &pt) {
    // TODO
}

void api_driver::ApiDriver::setNetworkSettings(boost::property_tree::ptree &pt) {
    // TODO
}

void api_driver::ApiDriver::resetDefaultSettings() {
    daemon->resetDefaultSettings();
}

void api_driver::ApiDriver::executeInApi(const std::function<void(proxy::CpProxy&)>& callback) {
    try {
        std::lock_guard lock(this->daemon->cpApiMutex);
        callback(this->daemon->cp);
    } catch (std::exception& e) {
        BOOST_LOG_TRIVIAL(error) << "ApiDriver::executeInApi(): failed to exec with error: " << e.what();
    }
}

#ifdef MODEM_IS_TDMA
std::string api_driver::ApiDriver::getOtaFileLocation() const {
    obj::TerminalDeviceState ds;
    daemon->getDeviceState(ds);
    return ds.fUpgradeImage;
}
#endif

#ifdef MODEM_IS_SCPC
std::string api_driver::ApiDriver::getLoggingStatisticsSettings() {
    return this->daemon->statsLogs.getSettings();
}

void api_driver::ApiDriver::setLoggingStatisticsSettings(boost::property_tree::ptree &pt) {
    this->daemon->statsLogs.setSettings(pt);
}
#endif

api_driver::ApiDriver::~ApiDriver() = default;
